/* Default config.h generated by Makefile */
#ifndef _XOPEN_SOURCE
#define _XOPEN_SOURCE 600
#endif
#define HAVE_LIBBZ2 1
#define HAVE_LIBLZMA 1
#ifndef __APPLE__
#define HAVE_LZMA_H 1
#endif
#define HAVE_LIBCURL 1
